#R Version: 4.0.2
#Alignator Version: 1.1

runFunc <- function() {
  # get directory of script and annotation files
  script.dir <- getSrcDirectory(function(dummy){dummy})
  wd <- getwd()
  
  # install necessary libraries for GUI
  if (!is.element("gWidgets2",installed.packages()[,1]))
  { 
    install.packages("gWidgets2")
    install.packages("RGtk2", depen=T)
    install.packages("gWidgets2RGtk2")
  }
  
  require(gWidgets2)
  require(tools)
  
  # Install Packages
  if (!"BiocManager" %in% rownames(installed.packages())){
    install.packages(c("BiocManager", "plyr", "tidyverse"))
    BiocManager::install(c("Rsubread","edgeR")) 
  }
  
  # Libraries
  require(Rsubread)
  require(edgeR)
  require(plyr)
  require(parallel)
  require(tidyverse)
  
  # empty list with references
  referenceNames <- c("No reference added yet")
  
  #check if config file with reference names exists. 
  if (file.exists(paste0(script.dir, "/references_config.txt"))) {
    config <- read.table(paste0(script.dir, "/references_config.txt"), sep = "\n")
    referenceNames <- config[[1]]
  }
  
  #handler functions
  addReference <- function(h, ...)
  {
    
    #open new window to chose the paths for a new reference
    enabled(addReference) <- F
    refW <- gwindow("add reference", visible = TRUE, height = 100, width = 200)
    refParent <- ggroup(cont = refW, horizontal = FALSE, expand =F)
    lyt <- glayout(cont = refW, spacing = 5)
    right <- c(1,0); left <- c(-1,0)
    lyt[1,1, anchor = right] <- " Index File:"
    index.edit <- lyt[1,2, anchor = left] <- gedit(script.dir)
    size(index.edit) <- c(100,30)
    lyt[1,3, anchor = left] <- gbutton("open", handler = function(h, ...) {
      index.fasta <- gfile(script.dir, initial.dir = script.dir, filter = list(".fa fasta files" = list(patterns = c("*.fa"))))
      svalue(index.edit) <- index.fasta
    })
    lyt[2,1, anchor = right] <- "Name:"
    lyt[2,2, anchor = left] <- gedit("Reference Name", cont = lyt)
    lyt[3,2, anchor = right] <- buttons <- ggroup(cont = lyt)
    
    #buttons adds the reference to the reference matrix
    gbutton("add",cont=buttons,handler=function(h,...) {
      refAdd <- sapply(lyt[,2], svalue)
      
      #check if paths are valid
      if (length(refAdd[2])==0) galert("no reference name given", parent =refW)
      
      #check index, build index file from .fa and make annotations from file:
      if (length(refAdd[[1]])!=0) {
        enabled(refW) <- F
        index.name <- paste0(refAdd[2], "_index")
        
        
        #create directory for storage of reference index and annotations
        dir.create(paste0(script.dir,"/",refAdd[2]))
        setwd(paste0(script.dir, "/", refAdd[2]))
        
        buildindex(basename= index.name,
                   reference = refAdd[1],
                   gappedIndex = F,
                   indexSplit = F,
                   memory= 8000,
                   TH_subread= 100)
        
        setwd(wd)
        
        #create annotations 
        makeAnnotations(refAdd[1], refAdd[2])
        
        enabled(refW) <- T
      }
      else galert("index file invalid", parent = refW)
      
      if ((length(refAdd[[1]]) != 0) && (length(refAdd[[2]]) != 0))
      {
        if (!file.exists(paste0(script.dir, "/references_config.txt"))) {
          referenceNames <<- refAdd[2]
          enabled(cbReference) <- T
          enabled(btnAlign) <- T
        } else referenceNames <<- append(referenceNames, refAdd[2])
        
        #add to references file
        write(refAdd[2],file=paste0(script.dir, "/references_config.txt"),append=TRUE)
        
        cbReference[] <- referenceNames
        svalue(cbReference, index = T) <- 1
        dispose(refW)
        enabled(addReference) <- T
      }
      
    } )
    gbutton("cancel",cont=buttons,handler=function(h,...) {
      dispose(refW)
      enabled(addReference) <- T
    } )
  }
  
  makeAnnotations <- function(input_path, ref_name) {
    #read in file:
    cat(noquote("### read in fasta file ###\n\n"))
    anno <- read.csv(input_path, header = F)
    
    relevant_lines <- startsWith(anno[,], ">")
    relevant_sum <- sum(relevant_lines)
    
    #matrix with all information:
    anno_matrix <- matrix(nrow = sum(relevant_lines), ncol = 12)
    colnames(anno_matrix) <- c("Transcript stable ID version", "Chromosome/scaffold", "Assembly", "Chromosome number", 
                               "Transcripton start", "Transcription end", "Transcription Direction", "Gene stable ID", "Gene Biotype",
                               "Transcript type","Gene name", "Gene description")
    
    relevant_num <- (1:length(relevant_lines))[relevant_lines]
    index <- 1
    progress <- 0
    
    #get all data from file
    cat(noquote("### create annotations file ###\n"))
    for (current_line in relevant_num){
      split_line <- strsplit(anno[current_line,], " ")
      anno_matrix[index, 1] <- gsub(">", "", split_line[[1]][1])
      # anno_matrix[index, 2] <- split_line[[1]][3]
      #get all information of chromosomes in separate columns
      chromosome_info <- strsplit(split_line[[1]][3], ":")[[1]]
      for (k in 1:length(chromosome_info)) {
        anno_matrix[index, k+1] <- chromosome_info[k]
      }
      anno_matrix[index, 8] <- gsub("gene:", "", split_line[[1]][4])
      anno_matrix[index, 9] <- gsub("gene_biotype:", "", split_line[[1]][5])
      anno_matrix[index, 10] <- gsub("transcript_biotype:", "", split_line[[1]][6])
      anno_matrix[index, 11] <- gsub("gene_symbol:", "", split_line[[1]][7])
      anno_matrix[index, 12] <- paste(gsub("description:", "", split_line[[1]][-(1:7)]), collapse = " ")
      #some progress here
      if (index > (relevant_sum/10)*progress){
        cat(noquote(paste0("### [", progress*10 , " % finished] ###\n")))
        progress <- progress +1
      }
      index <- index + 1
    }
    cat(noquote("### [100 % finished] ###\n\n"))
    
    #save annotations to script path
    cat(noquote("### save annotations file ###\n\n"))
    write.table(anno_matrix, paste0(script.dir,"/",ref_name, "/", ref_name, "_annotations.txt"), row.names = FALSE, quote = FALSE, sep = ",")
  }
  
  #create main window and groups for buttons
  options("guiToolkit"="RGtk2")
  w <- gwindow("Alignator v0-5",visible=FALSE, height = 190)
  
  parentCont <- ggroup(cont=w,horizontal=FALSE)
  frame <- gframe(cont=parentCont)
  
  mainLyt <- glayout(cont = frame, spacing = 5)
  right <- c(1,0); left <- c(-1,0)
  mainLyt[1,1, anchor = right] <- "output folder:"
  txtOutput <- mainLyt[1,2, anchor = left] <- gedit(script.dir)
  size(txtOutput) <- c(250,30)
  mainLyt[1,3, anchor = left] <- gbutton("open", handler = function(h,...) {
    txtOutputDirectory <- gfile(script.dir, initial.dir = script.dir,type="selectdir")
    svalue(txtOutput) <- txtOutputDirectory
  })
  mainLyt[2,1, anchor = right] <- " input files (.fastq):"
  txtInput <- mainLyt[2,2, anchor = left] <- gedit(script.dir)
  size(txtInput) <- c(250,30)
  mainLyt[2,3, anchor = left] <- gbutton("open", handler = function(h,...) {
    txtInputFile <<- gfile(script.dir,type="open", multi = T,initial.dir=script.dir, 
                           filter = list(".fastq files" = list(patterns = c("*.fastq"))))
    svalue(txtInput) <- txtInputFile[1]
  })
  mainLyt[3,1, anchor = right] <- "reference:"
  cbReference <- mainLyt[3,2, anchor = left] <- gcombobox(referenceNames)
  
  mainLyt[3,3, anchor = left] <- gbutton("add reference", handler = addReference)
  mainLyt[4,2, anchor = left] <- glabel()
  spaceLbl <- mainLyt[6,2, anchor = left] <- glabel()
  btnAlign <- mainLyt[5,2, anchor = left] <- gbutton("align", handler = function(h,...){
    if (length(svalue(txtOutput))!=0) outputDirectory <- svalue(mainLyt[1,2])
    if (length(txtInputFile)!=0) inputFile <- txtInputFile
    chosenReference <- svalue(mainLyt[3,2])#, index = T)
    if ((length(outputDirectory)!=0) && (length(txtInputFile)!=0)) {
      svalue(spaceLbl) <- "aligning... please wait"
      align(outputDirectory, inputFile, chosenReference)
    } 
  })
  if (!file.exists(paste0(script.dir, "/references_config.txt"))) {
    enabled(cbReference) <- F
    enabled(btnAlign) <- F
  }
  
  visible(w) <- TRUE
  
  align <- function(outputDirectory, inputFile, chosenReference)
  {
    
    ########## Path of Working Directory ########
    
    cat("### \nprocessing files:")
    for (i in 1:length(inputFile)) cat(inputFile[i], "\n")
    cat("###\n\n")
    
    ###########  Loading FASTQ File (all merged into one)   ##################
    
    ## merge .fastq files if more than one is chosen  
    if (length(inputFile) > 1){
      cat("### merging files ###\n\n")
      mergeFile <- paste0(file_path_sans_ext(inputFile[1]), "_merged.fastq")
      file.append(mergeFile, inputFile)
      inputFile <- mergeFile
    }
    
    ##########   Alignment Reads ###############
    cat("### creating SAM file ###\n\n")
    
    # index has to be in same folder as script
    sublong(
      index= paste0(script.dir, "/", chosenReference, "/", chosenReference, "_index"), #referenceMatrix[chosenReference, "Index"],
      readFiles= inputFile,
      outputFiles= paste(inputFile,"sublong","SAM",sep = "."),
      outputFormat = "SAM",
      nthreads = detectCores(all.tests = F, logical = T))
    
    SAMfile <- paste(inputFile,"sublong","SAM",sep = ".")
    
    ##############   Reading SAM File and Removal of @ rows   ##############
    
    cat("### aligning ###\n\n")
    omit<- c("@HD","@SQ","@PG")
    
    Column.name <- c("Read_ID","Flag","Transcript",
                     "Position","Mapping_Quality","CIGAR","RNEXT",
                     "PNEXT","TLEN","Sequence","Quality_Read",
                     "Optional_Field")
    
    SAM <- read_table2(SAMfile,
                       col_names = Column.name)
    
    reads <-SAM[!SAM$Read_ID %in% omit,]
    
    
    ############  Exclude not aligned reads #########
    
    Aligned <- reads[reads$Transcript != '*',]
    
    ############   Annotation of RNA Data with Annotation_File #############
    ############   Choose Annotation File   ####################################
    
    annotations <- read.csv(paste0(script.dir, "/", chosenReference, "/", chosenReference, "_annotations.txt"), # referenceMatrix[chosenReference, "Annotation"], 
                            header=TRUE,
                            na.strings = "") #choose file
    
    Annotated_Data <<- merge(Aligned,annotations,by.x="Transcript", by.y="Transcript.stable.ID.version", all.x=T)
    
    #############   Read Count  #######################
    
    Gene_Frequency <<- count(Annotated_Data, Gene.name)
    
    cat(noquote("### save Annotations and Counts as \"AnnotatedData.xls\" and \"GeneCounts.xls\" ###"))
    write.table(Annotated_Data, 
                file = paste0(outputDirectory, "/AnnotatedData.xls"), 
                sep = "\t",
                na= "NA", 
                row.names = TRUE, 
                col.names = NA, 
                quote = F)
    
    #############  Export Counts 
    write.table(Gene_Frequency, 
                file = paste0(outputDirectory, "/GeneCounts.xls"), 
                sep = "\t",
                row.names = FALSE)
    svalue(spaceLbl) <- "Done."
    
  }
}
runFunc()